"""Provides functionality to create networkx graphs and pltoly figures for network visualization"""

import networkx as nx
import pandas as pd

LAYOUT_SEED = 7


def initalize_network(
    interactions: pd.DataFrame, terms: list, search_mode: str
) -> nx.Graph:
    """Create a networkx graph representing interactions between genes and drugs

    :param interactions: DataFrame containing drug-gene interaction data
    :param terms: List containing terms used to query interaction data
    :param search_mode: String indicating whether query was gene-focused or drug-focused
    :return: a networkx graph of drug-gene interactions
    """
    interactions_graph = nx.Graph()
    graphed_terms = set()

    for index in range(len(interactions["gene_name"]) - 1):
        if search_mode == "genes":
            graphed_terms.add(interactions["gene_name"][index])
        if search_mode == "drugs":
            graphed_terms.add(interactions["drug_name"][index])
        interactions_graph.add_node(
            interactions["gene_name"][index],
            label=interactions["gene_name"][index],
            isGene=True,
        )
        interactions_graph.add_node(
            interactions["drug_name"][index],
            label=interactions["drug_name"][index],
            isGene=False,
        )
        interactions_graph.add_edge(
            interactions["gene_name"][index],
            interactions["drug_name"][index],
            id=interactions["gene_name"][index]
            + " - "
            + interactions["drug_name"][index],
            approval=interactions["drug_approved"][index],
            score=interactions["interaction_score"][index],
            attributes=interactions["interaction_attributes"][index],
            sourcedata=interactions["interaction_sources"][index],
            pmid=interactions["interaction_pmids"][index],
        )

    graphed_terms = set(terms).difference(graphed_terms)
    for term in graphed_terms:
        if search_mode == "genes":
            interactions_graph.add_node(term, label=term, isGene=True)
        if search_mode == "drugs":
            interactions_graph.add_node(term, label=term, isGene=False)

    nx.set_node_attributes(
        interactions_graph, dict(interactions_graph.degree()), "node_degree"
    )
    return interactions_graph


def _add_node_attributes(interactions_graph: nx.Graph, search_mode: str) -> None:
    for node in interactions_graph.nodes:
        is_gene = interactions_graph.nodes[node]["isGene"]
        degree = interactions_graph.degree[node]
        if search_mode == "genes":
            if is_gene:
                if degree > 1:
                    set_color = "cyan"
                    set_size = 10
                else:
                    set_color = "blue"
                    set_size = 10
            else:
                if degree > 1:
                    set_color = "orange"
                    set_size = 7
                else:
                    set_color = "red"
                    set_size = 7
        if search_mode == "drugs":
            if is_gene:
                if degree > 1:
                    set_color = "cyan"
                    set_size = 7
                else:
                    set_color = "blue"
                    set_size = 7
            else:
                if degree > 1:
                    set_color = "orange"
                    set_size = 10
                else:
                    set_color = "red"
                    set_size = 10
        interactions_graph.nodes[node]["node_color"] = set_color
        interactions_graph.nodes[node]["node_size"] = set_size


def create_network(
    interactions: pd.DataFrame, terms: list, search_mode: str
) -> nx.Graph:
    """Create a networkx graph representing interactions between genes and drugs

    :param interactions: DataFrame containing drug-gene interaction data
    :param terms: List containing terms used to query interaction data
    :param search_mode: String indicating whether query was gene-focused or drug-focused
    :return: a networkx graph of drug-gene interactions
    """
    interactions_graph = initalize_network(interactions, terms, search_mode)
    _add_node_attributes(interactions_graph, search_mode)
    return interactions_graph


def generate_cytoscape(graph: nx.Graph) -> dict:
    """Create a cytoscape graph representing interactions between genes and drugs

    :param graph: networkx graph to be formatted as a cytoscape graph
    :return: a cytoscape graph of drug-gene interactions
    """
    pos = nx.spring_layout(graph, seed=LAYOUT_SEED, scale=4000)
    cytoscape_data = nx.cytoscape_data(graph)["elements"]
    cytoscape_node_data = cytoscape_data["nodes"]
    cytoscape_edge_data = cytoscape_data["edges"]
    for node in range(len(cytoscape_node_data)):
        node_pos = pos[cytoscape_node_data[node]["data"]["id"]]
        node_pos = {
            "position": {"x": int(node_pos[0].item()), "y": int(node_pos[1].item())}
        }
        cytoscape_node_data[node].update(node_pos)
    return cytoscape_node_data + cytoscape_edge_data



"""Provide basic test configuration and fixture root."""

from io import TextIOWrapper
from pathlib import Path

import pytest
import requests_mock


def pytest_addoption(parser):
    parser.addoption("--performance", action="store_true", help="Run performance tests")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--performance"):
        return
    skip_custom = pytest.mark.skip(reason="Need --performance option to run")
    for item in items:
        if "performance" in item.keywords:
            item.add_marker(skip_custom)


@pytest.fixture(scope="session")
def fixtures_dir():
    """Provide path to fixtures directory."""
    return Path(__file__).resolve().parent / "fixtures"


@pytest.fixture(scope="session")
def schema(fixtures_dir):
    with (fixtures_dir / "introspection_response.json").open() as f:
        return f.read()


@pytest.fixture(scope="session")
def set_up_graphql_mock(schema: str):
    def _set_up_graphql_mock(m: requests_mock.Mocker, json_response: TextIOWrapper):
        """Initialize mock for a new set of GraphQL requests.

        The client will first ping the server for a schema, and then send a request that
        has been validated against that schema. This method ensures that Mockers are called
        in the correct way to handle this (counterintuitively, we first set the test-specific
        response and then, secondly, add a listener with a custom match pattern for the
        schema response).

        :param m: mock requests object
        :param schema_response: schema description for introspection/query validation
        :param json_response: expected query response from the server
        """
        m.post(
            "https://dgidb.org/api/graphql",
            text=json_response.read(),
        )
        m.post(
            "https://dgidb.org/api/graphql",
            additional_matcher=lambda r: "IntrospectionQuery"
            in r.json().get("query", ""),
            text=schema,
        )

    return _set_up_graphql_mock



"""Test `dgipy.network.construct`."""

import json
from pathlib import Path

from dgipy.network import construct


def test_construct(fixtures_dir: Path):
    results = json.load(
        (fixtures_dir / "construct_network_input_interactions.json").open()
    )
    graph = construct.construct_graph(results)
    assert len(graph.nodes) == 7
    assert len(graph.edges) == 6
    assert graph.nodes["hgnc:3443"]["long_name"] == "epiregulin"
    assert graph.nodes["ncit:C188574"]["approved"] is False
    assert graph.edges[("hgnc:3443", "iuphar.ligand:7836")]["attributes"][
        "Mechanism of Action"
    ] == ["Inhibition"]



"""Define methods for exporting to common knowledge graph frameworks."""

import networkx as nx


def to_pykeen(graph: nx.Graph) -> list[tuple[str, str, str]]:
    """Export to PyKEEN triple set. Typically, you'd save this output to a TSV.

    PyKEEN likes very straightforward triples. There's probably work we could do to
    better characterize the interaction, and also add attributes as additional triples.
    As it stands, this method is VERY basic.

    :param graph: graph constructed from DGIpy results.
    :return: list of triples (e.g. to save to TSV)
    """
    triples = []
    for gene_id, drug_id in graph.edges:
        gene = graph.nodes[gene_id]["name"]
        drug = graph.nodes[drug_id]["name"]
        triples.append((gene, "has_drug_target_interaction_with", drug))
        triples.append((drug, "has_drug_target_interaction_with", gene))

    return triples



"""Python wrapper for accessing an instance of DGIdb v5 database"""

from .dgidb import (
    SourceType,
    get_all_genes,
    get_categories,
    get_drug_applications,
    get_drugs,
    get_genes,
    get_interactions,
    get_sources,
)
from .graph_app import generate_app

__all__ = [
    "get_drugs",
    "get_genes",
    "get_interactions",
    "get_categories",
    "get_sources",
    "SourceType",
    "get_all_genes",
    "get_drug_applications",
    "generate_app",
]



"""Provide queries and lazy accessors to queries.


Individual query loader classes provide ``gql`` queries via a ``.query`` property:

>>> from dgipy.queries import get_drugs
>>> get_drugs.query
DocumentNode at 0:545
"""

from importlib import resources

from gql import gql
from graphql import DocumentNode


class _LazyQueryLoader:
    """Provide lazy loading functionality for query access."""

    def __init__(self, query_name: str) -> None:
        self.query_name = query_name
        self._query = None

    @property
    def query(self) -> DocumentNode:
        if self._query is None:
            with resources.open_text(__name__, f"{self.query_name}.graphql") as f:
                self._query = gql(f.read())
        return self._query


get_all_genes = _LazyQueryLoader("get_all_genes")
get_all_drugs = _LazyQueryLoader("get_all_drugs")
get_drug_applications = _LazyQueryLoader("get_drug_applications")
get_drugs = _LazyQueryLoader("get_drugs")
get_gene_categories = _LazyQueryLoader("get_gene_categories")
get_genes = _LazyQueryLoader("get_genes")
get_interactions_by_drug = _LazyQueryLoader("get_interactions_by_drug")
get_interactions_by_gene = _LazyQueryLoader("get_interactions_by_gene")
get_sources = _LazyQueryLoader("get_sources")


__all__ = [
    "get_all_genes",
    "get_all_drugs",
    "get_drug_applications",
    "get_drugs",
    "get_gene_categories",
    "get_genes",
    "get_interactions_by_drug",
    "get_interactions_by_gene",
    "get_sources",
]



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "DGIpy"
copyright = "2024, Wagner Lab"
author = "Wagner Lab"
html_title = "DGIpy"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
    "sphinx.ext.linkcode",
    "sphinx_copybutton",
    "sphinx_github_changelog",
]

templates_path = ["_templates"]
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = []
html_css_files = [
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.0.0/css/fontawesome.min.css",
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.0.0/css/solid.min.css",
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.0.0/css/brands.min.css",
]
html_theme_options = {
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/genomicmedlab/dgipy",
            "html": "",
            "class": "fa-brands fa-solid fa-github",
        },
        {
            "name": "Wagner Lab",
            "url": "https://www.nationwidechildrens.org/specialties/institute-for-genomic-medicine/research-labs/wagner-lab",
            "html": "",
            "class": "fa-solid fa-house",
        }
    ],
}
# -- autodoc things ----------------------------------------------------------
import os  # noqa: E402
import sys  # noqa: E402

sys.path.insert(0, os.path.abspath("../../dgipy"))
autodoc_preserve_defaults = True

# -- get version -------------------------------------------------------------
from importlib.metadata import version

release = version = version("dgipy")

# -- linkcode ----------------------------------------------------------------
def linkcode_resolve(domain, info):
    if domain != "py":
        return None
    if not info["module"]:
        return None
    filename = info["module"].replace(".", "/")
    return f"https://github.com/genomicmedlab/dgipy/blob/main/{filename}.py"


# -- code block style --------------------------------------------------------
pygments_style = "default"
pygements_dark_style = "monokai"



"""Construct a NetworkX graph from DGIpy query results."""

import networkx as nx

from dgipy.data_utils import make_tabular


def _get_gene_nodes(result_table: list[dict]) -> list[tuple[str, dict]]:
    if result_table and "gene_concept_id" not in result_table[0]:
        return []

    nodes = []
    for row in result_table:
        node_attrs = {"type": "gene"}
        node_attrs.update(
            {
                k[5:]: v
                for k, v in row.items()
                if k.startswith("gene_") and not k.startswith("gene_concept_id")
            }
        )
        nodes.append((row["gene_concept_id"], node_attrs))
    return nodes


def _get_drug_nodes(result_table: list[dict]) -> list[tuple[str, dict]]:
    if result_table and "drug_concept_id" not in result_table[0]:
        return []

    nodes = []
    for row in result_table:
        node_attrs = {"type": "drug"}
        node_attrs.update(
            {
                k[5:]: v
                for k, v in row.items()
                if k.startswith("drug_") and not k.startswith("drug_concept_id")
            }
        )
        nodes.append((row["drug_concept_id"], node_attrs))
    return nodes


def _get_interaction_edges(result_table: list[dict]) -> list[tuple[str, str, dict]]:
    if result_table and (
        "drug_concept_id" not in result_table[0]
        or "gene_concept_id" not in result_table[0]
    ):
        return []

    edges = []
    for row in result_table:
        edge_attrs = {"type": "drug_gene_interaction"}
        edge_attrs.update(
            {k[12:]: v for k, v in row.items() if k.startswith("interaction_")}
        )
        edges.append((row["gene_concept_id"], row["drug_concept_id"], edge_attrs))
    return edges


def construct_graph(query_result: dict) -> nx.Graph:
    """Construct a NetworkX graph from a DGIpy query result (i.e., a columnar dict).

    >>> import dgipy
    ... from dgipy.network.construct import construct_graph
    >>> genes = dgipy.get_genes(["BRAF", "ABL1"])
    >>> graph = construct_graph(genes)

    :param query_result: result object directly from DGIpy output. In general, columns
        with names starting with ``"drug_"`` will be added as attributes of drug nodes,
        ``"gene_"`` (excluding ``"gene_category_"``) as attributes of gene nodes, and
        ``"interaction"`` as part of interaction edges.
    :return: nx.Graph, where any included drug, gene, and gene category instances are
        nodes, and edges are drawn between interacting genes and drugs, as well as
        genes and their corresponding gene categories.
    """
    graph = nx.Graph()
    result_table = make_tabular(query_result)

    graph.add_nodes_from(_get_gene_nodes(result_table))
    graph.add_nodes_from(_get_drug_nodes(result_table))
    graph.add_edges_from(_get_interaction_edges(result_table))

    return graph



"""Provide tools for network-based operations.

For our purposes, drugs and genes are the only nodes on the network, because they're the
only real "entities". Groupings like categories and attributes are defined as properties
of the nodes themselves, but different kinds of processing methods and frameworks might
want to break them out as distinct nodes that relate logically to drugs and genes.
"""



"""Provides functionality to create a Dash web application for interacting with drug-gene data from DGIdb"""

import json

import dash_bootstrap_components as dbc
import dash_cytoscape as cyto
from dash import Input, Output, State, ctx, dash, dcc, html

from dgipy import dgidb
from dgipy import network_graph as ng
from dgipy.data_utils import make_tabular

cyto.load_extra_layouts()


def generate_app() -> dash.Dash:
    """Initialize a Dash application object with a layout designed for visualizing: drug-gene interactions, options for user interactivity, and other visual elements.

    :return: a python dash app that can be run with run_server()
    """
    genes = [
        {"label": gene["gene_name"], "value": gene["gene_name"]}
        for gene in make_tabular(dgidb.get_all_genes())
    ]
    drugs = [
        {"label": drug["drug_name"], "value": drug["drug_name"]}
        for drug in make_tabular(dgidb.get_all_drugs())
    ]

    app = dash.Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])

    _set_app_layout(app)
    _update_cytoscape(app)
    _update_terms_dropdown(app, genes, drugs)
    _update_selected_element(app)
    _update_selected_element_text(app)
    _update_neighbors_dropdown(app)
    _update_edge_info(app)
    _generate_image(app)
    _generate_json(app)

    return app


def _set_app_layout(app: dash.Dash) -> None:
    cytoscape_figure = cyto.Cytoscape(
        id="cytoscape-figure",
        layout={"name": "preset"},
        style={"width": "100%", "height": "800px"},
        stylesheet=[
            # Group Selectors
            {
                "selector": "node",
                "style": {"content": "data(label)"},
            },
            {"selector": "edge", "style": {"width": 0.75}},
            {
                "selector": "[node_degree > 1][isGene]",
                "style": {
                    "background-color": "cyan",
                },
            },
            {
                "selector": "[node_degree <= 1][isGene]",
                "style": {
                    "background-color": "blue",
                },
            },
            {
                "selector": "[node_degree > 1][!isGene]",
                "style": {
                    "background-color": "orange",
                },
            },
            {
                "selector": "[node_degree <= 1][!isGene]",
                "style": {
                    "background-color": "red",
                },
            },
        ],
    )

    search_mode = dcc.RadioItems(
        id="search-mode",
        options=[
            {"label": "Gene Search", "value": "genes"},
            {"label": "Drug Search", "value": "drugs"},
        ],
        value="genes",
    )

    terms_dropdown = dcc.Dropdown(
        id="terms-dropdown", optionHeight=75, multi=True, value=[]
    )

    selected_element_text = dcc.Markdown(
        id="selected-element-text", children="No Element Selected"
    )

    neighbors_dropdown = dcc.Dropdown(id="neighbors-dropdown", multi=False)

    selected_edge_info = dcc.Markdown(
        id="selected-edge-info", children="No Edge Selected"
    )

    app.layout = html.Div(
        [
            # Variables
            dcc.Store(id="selected-element", data=""),
            dcc.Store(id="graph"),
            # Layout
            dbc.Row(
                [
                    dbc.Col(
                        dbc.Card(cytoscape_figure, body=True, style={"margin": "10px"}),
                        width=8,
                    ),
                    dbc.Col(
                        [
                            dbc.Card(
                                [
                                    dbc.CardHeader("Search Mode"),
                                    dbc.CardBody(search_mode),
                                ],
                                style={"margin": "10px"},
                            ),
                            dbc.Card(
                                [
                                    dbc.CardHeader("Terms Dropdown"),
                                    dbc.CardBody(terms_dropdown),
                                ],
                                style={"margin": "10px"},
                            ),
                            dbc.Card(
                                [
                                    dbc.CardHeader("Neighbors Dropdown"),
                                    dbc.CardBody(neighbors_dropdown),
                                ],
                                style={"margin": "10px"},
                            ),
                            dbc.Card(
                                [
                                    dbc.CardHeader("Selection Info"),
                                    dbc.CardBody(
                                        [
                                            html.H4("Selected Node/Edge:"),
                                            html.P(selected_element_text),
                                            html.H4("Selected Edge Info:"),
                                            html.P(selected_edge_info),
                                        ]
                                    ),
                                ],
                                style={"margin": "10px"},
                            ),
                            dbc.Card(
                                [
                                    dbc.CardHeader("Export Graph"),
                                    dbc.CardBody(
                                        [
                                            dbc.Button(
                                                "Export Graph as .png",
                                                id="export-png-graph",
                                                class_name="m-1",
                                            ),
                                            dbc.Button(
                                                "Export Graph as .svg",
                                                id="export-svg-graph",
                                                class_name="m-1",
                                            ),
                                            dbc.Button(
                                                "Export Graph as .json",
                                                id="export-json-graph",
                                                class_name="m-1",
                                            ),
                                            dcc.Download(id="json-download"),
                                        ]
                                    ),
                                ],
                                style={"margin": "10px"},
                            ),
                        ],
                        width=4,
                    ),
                ]
            ),
        ]
    )


def _update_cytoscape(app: dash.Dash) -> None:
    @app.callback(
        Output("cytoscape-figure", "elements"),
        Input("terms-dropdown", "value"),
        State("search-mode", "value"),
    )
    def update(terms: list | None, search_mode: str) -> dict:
        if len(terms) != 0:
            interactions = dgidb.get_interactions(terms, search_mode)
            network_graph = ng.initalize_network(interactions, terms, search_mode)
            return ng.generate_cytoscape(network_graph)
        return {}


def _update_terms_dropdown(app: dash.Dash, genes: list, drugs: list) -> None:
    @app.callback(
        Output("terms-dropdown", "options"),
        Input("search-mode", "value"),
    )
    def update(search_mode: str) -> list:
        if search_mode == "genes":
            return genes
        if search_mode == "drugs":
            return drugs
        return None


def _update_selected_element(app: dash.Dash) -> None:
    @app.callback(
        Output("selected-element", "data"),
        [
            Input("cytoscape-figure", "tapNode"),
            Input("cytoscape-figure", "tapEdge"),
            Input("terms-dropdown", "value"),
        ],
    )
    def update(
        tap_node: dict | None,
        tap_edge: dict | None,
        terms_dropdown: list | None,  # noqa: ARG001
    ) -> str | dict:
        if ctx.triggered_prop_ids:
            dash_trigger = next(iter(ctx.triggered_prop_ids.keys()))
            if dash_trigger == "terms-dropdown.value":
                return ""
            if dash_trigger == "cytoscape-figure.tapNode" and tap_node is not None:
                return tap_node
            if dash_trigger == "cytoscape-figure.tapEdge" and tap_edge is not None:
                return tap_edge
        return dash.no_update


def _update_selected_element_text(app: dash.Dash) -> None:
    @app.callback(
        Output("selected-element-text", "children"), Input("selected-element", "data")
    )
    def update(selected_element: str | dict) -> str:
        if selected_element != "":
            return selected_element["data"]["id"]
        return "No Node Selected"


def _update_neighbors_dropdown(app: dash.Dash) -> None:
    @app.callback(
        [
            Output("neighbors-dropdown", "options"),
            Output("neighbors-dropdown", "value"),
        ],
        Input("selected-element", "data"),
    )
    def update(selected_element: str | dict) -> tuple[list, None]:
        if (
            selected_element != ""
            and selected_element["group"] == "nodes"
            and selected_element["data"]["node_degree"] != 1
        ):
            neighbor_set = set()
            for edge in selected_element["edgesData"]:
                neighbor_set.add(edge["target"])
                neighbor_set.add(edge["source"])
                neighbor_set.remove(selected_element["data"]["id"])
            neighbor_list = list(neighbor_set)
            return neighbor_list, None
        return [], None


def _update_edge_info(app: dash.Dash) -> None:
    @app.callback(
        Output("selected-edge-info", "children"),
        [Input("selected-element", "data"), Input("neighbors-dropdown", "value")],
    )
    def update(selected_element: str | dict, selected_neighbor: str | None) -> str:
        if selected_element == "":
            return "No Edge Selected"

        edge_info = None
        if selected_element["group"] == "nodes" and selected_neighbor is not None:
            edge_name = None
            if selected_element["data"]["isGene"]:
                edge_name = selected_element["data"]["id"] + " - " + selected_neighbor
            else:
                edge_name = selected_neighbor + " - " + selected_element["data"]["id"]
            for edge in selected_element["edgesData"]:
                if edge["id"] == edge_name:
                    edge_info = edge
                    break
        if selected_element["group"] == "edges":
            edge_info = selected_element["data"]

        if (
            selected_element["group"] == "nodes" and selected_neighbor is not None
        ) or selected_element["group"] == "edges":
            return (
                "ID: "
                + str(edge_info["id"])
                + "\n\nApproval: "
                + str(edge_info["approval"])
                + "\n\nScore: "
                + str(edge_info["score"])
                + "\n\nAttributes: "
                + str(edge_info["attributes"])
                + "\n\nSource: "
                + str(edge_info["source"])
                + "\n\nPmid: "
                + str(edge_info["pmid"])
            )
        return "No Edge Selected"


def _generate_image(app: dash.Dash) -> None:
    @app.callback(
        Output("cytoscape-figure", "generateImage"),
        [Input("export-png-graph", "n_clicks"), Input("export-svg-graph", "n_clicks")],
    )
    def update(export_png_graph: int, export_svg_graph: int) -> dict:  # noqa: ARG001
        if ctx.triggered_id == "export-png-graph":
            return {"type": "png", "action": "download"}
        if ctx.triggered_id == "export-svg-graph":
            return {"type": "svg", "action": "download"}
        return dash.no_update


def _generate_json(app: dash.Dash) -> None:
    @app.callback(
        Output("json-download", "data"),
        Input("export-json-graph", "n_clicks"),
        State("cytoscape-figure", "elements"),
    )
    def update(export_png_graph: int, cytoscape_figure: dict) -> dict:  # noqa: ARG001
        if ctx.triggered_id is None:
            return dash.no_update
        return dcc.send_string(json.dumps(cytoscape_figure, indent=4), "cyto.json")



"""Provides methods for annotating VCF with DGIdb data"""

import contextlib
from itertools import groupby
from pathlib import Path

import pandas as pd
import pysam
import requests
from tqdm import tqdm

import dgipy


# TODO: Probably need another class as a wrapper object rather than putting it all in a list
# Class would have analogous display methods but also allow access to individual GeneResults
class GeneResult:
    """A gene result from original VCF
    .. code-block:: python
        import vcf

        # Provide path to VCF file and specify chromosome
        data = vcf.annotate("link/to/file", chr="chr#")
    """

    def __init__(self, data: list) -> None:
        """Initialize a gene with a name, VCF records, and drug-gene interactions"""
        # TODO: handle genes without names, 'novel transript'
        try:
            self.gene = data[0]["name"]
        except:
            self.gene = "None"

        self.records = data
        self.interactions = dgipy.get_interactions(self.gene)

        # TODO: handle app searches with blank lists [], currently FDA resource hangs for awhile?
        if not list(self.interactions["drug"].values):
            self.applications = "None"
        else:
            self.applications = dgipy.get_drug_applications(
                list(self.interactions["drug"].values)
            )

        self.gene_info = dgipy.get_genes(self.gene)
        self.categories = dgipy.get_categories(self.gene)


def annotate(filepath: Path, contig: str) -> pd.DataFrame:
    """Map chr,pos pairs from a VCF file to human genes and search DGIdb for drug-gene interactions

    :param filepath: link to a valid VCF file
    :param contig: specified chromosome (i.e. chr7)
    :return: Dataframe of drug-gene interactions
    """
    if not isinstance(filepath, Path):
        msg = "Filepath argument must be a valid pathlib.Path object"
        raise ValueError(msg)

    # Open VCF file
    records = _process_vcf(filepath, contig)
    # Grab records & relevant info (params: chr7)
    mapped = _ensembl_map(records)  # TODO: modularize mapping
    # Group records with like-genes
    grouped = _group_by_name(mapped)
    # Instance each gene set as a class
    return [GeneResult(grouped[gene]) for gene in grouped]


def _process_vcf(filepath: Path, contig: str) -> list:
    """Grab relevant data for mapping and mutations from starting VCF

    :param filepath: link to valid VCF file
    :param contig: specified chromosome (i.e. chr7)
    :return: List of record dicts

    """
    # TODO: Add support for pysam.VariantFile
    # https://pysam.readthedocs.io/en/latest/usage.html#working-with-vcf-bcf-formatted-files

    file = pysam.TabixFile(str(filepath.absolute()))

    records = []
    for record in tqdm(file.fetch(contig)):
        fields = record.split("\t")
        entry = {
            "chromosome": fields[0],
            "pos": fields[1],
            "ref": fields[3],
            "alt": fields[4],
            "qual": fields[5],
            "filter": fields[6],
        }
        records.append(entry)

    return records


def _get_gene_by_position(chromosome: str, position: str) -> list:
    """Map chr,pos pair to genome via ensembl

    :param chromosome: specified chromosome (i.e. chr7)
    :param position: genomic coordinate
    :return: genomic info for specified coordinate
    """
    url = f"https://rest.ensembl.org/overlap/region/human/{chromosome}:{position}-{position}?feature=gene"
    headers = {"Content-Type": "application/json"}
    response = requests.get(f"{url}", headers=headers, timeout=10)

    if not response.ok:
        response.raise_for_status()
        return None

    return response.json()


def _ensembl_map(records: list) -> list:
    """Take VCF input and map to ensembl

    :param records: list of records pulled from VCF
    :return: list of mapped genes
    """
    results = []
    # TODO: Allow custom slice selection as data sets can be huge, currently slicing 0:1500 or 0:150 for time purposes
    for record in tqdm(records[0:1500]):
        gene_info = _get_gene_by_position(record["chromosome"], record["pos"])

        if type(gene_info) is None:
            continue

        for info in gene_info:
            entry = {}
            if info["feature_type"] == "gene":
                with contextlib.suppress(KeyError):  # TODO: handle genes without names
                    entry["name"] = info["external_name"]

                entry["desc"] = info["description"]
                entry["gene_id"] = info["gene_id"]
                entry.update(record)
                results.append(entry)

    return results


def _group_by_name(data: list, default_name: str = "Unknown") -> dict:
    """Take list of records and group to dict by gene name

    :param data: list of records
    :param default_name: name of gene if none found
    :return: dict of records grouped by gene name
    """
    sorted_data = sorted(data, key=lambda x: x.get("name", default_name))

    return {
        key: list(group)
        for key, group in groupby(
            sorted_data, key=lambda x: x.get("name", default_name)
        )
    }



"""Integrate data from FDA clinical trials API."""

import logging

from regbot.fetch.clinical_trials import StandardAge, Status, Study
from regbot.fetch.clinical_trials import get_clinical_trials as get_trials_from_fda

_logger = logging.getLogger(__name__)


def _add_study_to_output(output: dict[str, list], drug_name: str, study: Study) -> None:
    """Update `output` in-place with results from study

    :param output: in-progress raw columnar data
    :param drug_name: name of drug that was searched
    :param study: clinical trial study data to add to output
    """
    output["drug_name"].append(drug_name.upper())
    output["trial_id"].append(study.protocol.identification.nct_id)
    output["brief"].append(study.protocol.identification.brief_title)
    output["study_type"].append(study.protocol.design.study_type)
    min_age = (
        study.protocol.eligibility.min_age
        if study.protocol and study.protocol.eligibility
        else None
    )
    output["min_age"].append(min_age)
    max_age = (
        study.protocol.eligibility.max_age
        if study.protocol and study.protocol.eligibility
        else None
    )
    output["max_age"].append(max_age)
    age_groups = (
        study.protocol.eligibility.std_age
        if study.protocol and study.protocol.eligibility
        else None
    )
    output["age_groups"].append(age_groups)
    output["pediatric"].append(StandardAge.CHILD in age_groups if age_groups else None)
    output["conditions"].append(
        study.protocol.conditions.conditions
        if study.protocol and study.protocol.conditions
        else None
    )
    output["interventions"].append(
        [i._asdict() for i in study.protocol.arms_intervention.interventions]
        if study.protocol
        and study.protocol.arms_intervention
        and study.protocol.arms_intervention.interventions
        else None
    )
    eligibility = study.protocol.eligibility
    if not eligibility:
        output["incl_excl_criteria"].append(None)
        output["population_sex"].append(None)
        output["population_description"]
    else:
        output["incl_excl_criteria"].append(eligibility.description)
        output["population_sex"].append(eligibility.sex)
        output["population_description"].append(eligibility.population)
    all_locations = (
        study.protocol.contacts_locations.locations
        if study.protocol.contacts_locations
        and study.protocol.contacts_locations.locations
        else []
    )

    potential_sites = [
        {
            "name": location.facility,
            "status": location.status,
            "city": location.city,
            "country": location.country,
            "coordinates": location.geo,
        }
        for location in all_locations
        if location.status
        in {
            Status.RECRUITING,
            Status.NOT_YET_RECRUITING,
            Status.AVAILABLE,
            Status.TEMPORARILY_NOT_AVAILABLE,
            Status.UNKNOWN,
        }
    ]
    output["potential_sites"].append(potential_sites)


def get_clinical_trials(terms: list[str]) -> dict:
    """Acquire associated clinical trials data for drug term

    >>> from dgipy.dgidb import get_drugs
    >>> from dgipy.integration.clinical_trials import get_clinical_trials
    >>> import polars as pl  # or another dataframe library of your choosing
    >>> drugs = ["imatinib", "sunitinib"]
    >>> df = pl.DataFrame(get_drugs(drugs))
    >>> trial_df = pl.DataFrame(get_clinical_trials(drugs))
    >>> annotated_df = df.join(trial_df, on="drug_name")

    :param terms: drugs of interest
    :return: all clinical trials data for drugs of interest in a DataFrame-ready dict
    """
    if not isinstance(terms, list):
        _logger.warning(
            "Given `terms` arg doesn't appear to be a list. This argument should be a sequence of drug names (as strings)."
        )
    if not terms:
        msg = "Must supply nonempty argument for `terms`"
        raise ValueError(msg)

    output = {
        "drug_name": [],
        "trial_id": [],
        "brief": [],
        "study_type": [],
        "min_age": [],
        "max_age": [],
        "age_groups": [],
        "pediatric": [],
        "conditions": [],
        "interventions": [],
        "incl_excl_criteria": [],
        "population_sex": [],
        "population_description": [],
        "potential_sites": [],
    }

    for drug in terms:
        results = get_trials_from_fda(drug)

        for study in results:
            _add_study_to_output(output, drug, study)

    return output



import datetime
from pathlib import Path

import requests_mock
from regbot.fetch.clinical_trials import (
    InterventionType,
    StandardAge,
    Status,
    StudyType,
)

from dgipy.integrations.clinical_trials import get_clinical_trials


def test_get_clinical_trials(fixtures_dir: Path):
    with (
        requests_mock.Mocker() as m,
        (
            fixtures_dir / "integration_clinical_trials_zolgensma.json"
        ).open() as json_response,
    ):
        m.get(
            "https://clinicaltrials.gov/api/v2/studies?query.intr=zolgensma",
            text=json_response.read(),
        )
        results = get_clinical_trials(["zolgensma"])
        assert len(results["drug_name"]) == 18
        assert set(results["trial_id"]) == {
            "clinicaltrials:NCT06532474",
            "clinicaltrials:NCT04851873",
            "clinicaltrials:NCT05386680",
            "clinicaltrials:NCT04042025",
            "clinicaltrials:NCT04174157",
            "clinicaltrials:NCT03381729",
            "clinicaltrials:NCT05335876",
            "clinicaltrials:NCT03461289",
            "clinicaltrials:NCT03955679",
            "clinicaltrials:NCT03837184",
            "clinicaltrials:NCT03505099",
            "clinicaltrials:NCT02122952",
            "clinicaltrials:NCT06019637",
            "clinicaltrials:NCT05089656",
            "clinicaltrials:NCT05575011",
            "clinicaltrials:NCT05073133",
            "clinicaltrials:NCT03306277",
            "clinicaltrials:NCT03421977",
        }

        example_index = next(
            i
            for i, trial_id in enumerate(results["trial_id"])
            if trial_id == "clinicaltrials:NCT05386680"
        )
        assert (
            results["brief"][example_index]
            == "Phase IIIb, Open-label, Multi-center Study to Evaluate Safety, Tolerability and Efficacy of OAV101 Administered Intrathecally to Participants With SMA Who Discontinued Treatment With Nusinersen or Risdiplam"
        )
        assert results["study_type"][example_index] == StudyType.INTERVENTIONAL
        assert results["min_age"][example_index] == datetime.timedelta(days=730)
        assert results["age_groups"][example_index] == [StandardAge.CHILD]
        assert results["pediatric"][example_index] is True
        assert results["conditions"][example_index] == ["Spinal Muscular Atrophy"]
        assert results["interventions"][example_index] == [
            {
                "type": InterventionType.GENETIC,
                "name": "OAV101",
                "description": "Intrathecal administration of OAV101 at a dose of 1.2 x 10\\^14 vector genomes, one time dose",
                "aliases": ["AVXS-101", "Zolgensma"],
            }
        ]
        assert (
            results["incl_excl_criteria"][example_index]
            == "Inclusion Criteria\n\n* SMA diagnosis\n* Aged 2 to \\< 18 years\n* Have had at least four loading doses of nusinersen (Spinraza\u00ae) or at least 3 months of treatment with risdiplam (Evrysdi\u00ae) at Screening\n* Must have symptoms of SMA as defined in the protocol\n\nExclusion Criteria:\n\n* Anti Adeno Associated Virus Serotype 9 (AAV9) antibody titer using an immunoassay is reported as elevated\n* Clinically significant abnormalities in test results during screening\n* Contraindications for lumbar puncture procedure\n* At Baseline, participants are excluded if they received:\n\n  * nusinersen (Spinraza\u00ae) or\n  * risdiplam (Evrysdi\u00ae) within a defined timeframe\n* Vaccinations 2 weeks prior to administration of OAV101\n* Hospitalization for a pulmonary event, or for nutritional support within 2 months prior to Screening or inpatient major surgery planned.\n* Presence of an infection or febrile illness up to 30 days prior to administration of OAV101\n* Requiring invasive ventilation"
        )
        assert {
            "name": "Child Hosp of the Kings Daughters",
            "status": Status.RECRUITING,
            "city": "Norfolk",
            "country": "United States",
            "coordinates": (36.84681, -76.28522),
        } in results["potential_sites"][3]



from collections.abc import Callable
from io import StringIO
from pathlib import Path

import pytest
import requests_mock
from regbot.fetch.drugsfda import ProductDosageForm, ProductMarketingStatus

from dgipy.dgidb import (
    SourceType,
    get_all_genes,
    get_categories,
    get_drug_applications,
    get_drugs,
    get_genes,
    get_interactions,
    get_sources,
)


def test_get_drugs(fixtures_dir: Path, set_up_graphql_mock: Callable):
    with (
        requests_mock.Mocker() as m,
        (fixtures_dir / "get_drug_api_response.json").open() as json_response,
        (
            fixtures_dir / "get_drug_filtered_api_response.json"
        ).open() as filtered_json_response,
    ):
        set_up_graphql_mock(m, json_response)

        results = get_drugs(["Imatinib"])
        assert len(results["drug_name"]), "DataFrame is non-empty"

        results_with_added_fake = get_drugs(["imatinib", "not-real"])
        assert len(results_with_added_fake["drug_name"]) == len(
            results["drug_name"]
        ), "Gracefully ignore non-existent search terms"

        # handling filters
        filtered_results = get_drugs(["imatinib", "metronidazole"], antineoplastic=True)
        assert len(filtered_results["drug_name"]) == 1, "Metronidazole is filtered out"
        assert (
            filtered_results["drug_name"][0] == "IMATINIB"
        ), "Imatinib is retained by the filter"
        assert all(results["drug_is_antineoplastic"]), "All results are antineoplastics"

        set_up_graphql_mock(m, filtered_json_response)
        filtered_results = get_drugs(
            ["imatinib", "metronidazole"], antineoplastic=False
        )
        assert len(filtered_results["drug_name"]), "DataFrame is non-empty"
        assert "METRONIDAZOLE" in filtered_results["drug_name"]

        # empty response
        set_up_graphql_mock(m, StringIO('{"data": {"drugs": {"nodes": []}}}'))
        empty_results = get_drugs(["not-real"])
        assert len(empty_results["drug_name"]) == 0, "Handles empty response"


def test_get_genes(fixtures_dir: Path, set_up_graphql_mock: Callable):
    with (
        requests_mock.Mocker() as m,
        (fixtures_dir / "get_gene_api_response.json").open() as json_response,
    ):
        set_up_graphql_mock(m, json_response)

        results = get_genes(["ereg"])
        assert len(results["gene_name"]), "DataFrame is non-empty"

        results_with_added_fake = get_genes(["ereg", "not-real"])
        assert len(results_with_added_fake["gene_name"]) == len(
            results["gene_name"]
        ), "Gracefully ignore non-existent search terms"

        # empty response
        set_up_graphql_mock(m, StringIO('{"data": {"genes": {"nodes": []}}}'))
        empty_results = get_genes(["not-real"])
        assert len(empty_results["gene_name"]) == 0, "Handles empty response"


def test_get_interactions_by_genes(fixtures_dir: Path, set_up_graphql_mock: Callable):
    with (
        requests_mock.Mocker() as m,
        (
            fixtures_dir / "get_interactions_by_genes_response.json"
        ).open() as genes_response,
        (
            fixtures_dir / "get_interactions_by_multiple_genes_response.json"
        ).open() as multiple_genes_response,
    ):
        set_up_graphql_mock(m, genes_response)
        results = get_interactions(["ereg"])
        assert len(results["gene_name"]), "Results are non-empty"

        results = get_interactions(["ereg", "not-real"])
        assert len(results["gene_name"]), "Handles additional not-real terms gracefully"

        # multiple terms
        set_up_graphql_mock(m, multiple_genes_response)
        multiple_gene_results = get_interactions(["ereg", "braf"])
        assert len(multiple_gene_results["gene_name"]) > len(
            results["gene_name"]
        ), "Handles multiple genes at once"

        # empty response
        set_up_graphql_mock(m, StringIO('{"data": {"genes": {"nodes": []}}}'))
        empty_results = get_interactions(["not-real"])
        assert len(empty_results["gene_name"]) == 0, "Handles empty response"


def test_get_interactions_by_drugs(fixtures_dir: Path, set_up_graphql_mock: Callable):
    with (
        requests_mock.Mocker() as m,
        (
            fixtures_dir / "get_interactions_by_drugs_response.json"
        ).open() as drugs_response,
        (
            fixtures_dir / "get_interactions_by_multiple_drugs_response.json"
        ).open() as multiple_drugs_response,
    ):
        set_up_graphql_mock(m, drugs_response)
        results = get_interactions(["sunitinib"], search="drugs")
        assert len(results["drug_name"]), "Results are non-empty"

        results = get_interactions(["sunitinib", "not-real"], search="drugs")
        assert len(results["drug_name"]), "Handles additional not-real terms gracefully"

        # multiple terms
        set_up_graphql_mock(m, multiple_drugs_response)
        multiple_gene_results = get_interactions(
            ["sunitinib", "clonazepam"], search="drugs"
        )
        assert len(multiple_gene_results["drug_name"]) > len(
            results["drug_name"]
        ), "Handles multiple drugs at once"

        # empty response
        set_up_graphql_mock(m, StringIO('{"data": {"drugs": {"nodes": []}}}'))
        empty_results = get_interactions(["not-real"], search="drugs")
        assert len(empty_results["drug_name"]) == 0, "Handles empty response"


def test_get_categories(fixtures_dir: Path, set_up_graphql_mock: Callable):
    with (
        requests_mock.Mocker() as m,
        (fixtures_dir / "get_categories_response.json").open() as categories_response,
    ):
        set_up_graphql_mock(m, categories_response)
        results = get_categories(["BRAF"])
        assert len(results["gene_name"]), "Results are non-empty"
        assert "DRUG RESISTANCE" in results["gene_category"]
        assert "DRUGGABLE GENOME" in results["gene_category"]
        assert "CLINICALLY ACTIONABLE" in results["gene_category"]


def test_get_sources(fixtures_dir: Path, set_up_graphql_mock: Callable):
    with (
        requests_mock.Mocker() as m,
        (fixtures_dir / "get_sources_response.json").open() as sources_response,
        (
            fixtures_dir / "get_sources_filtered_response.json"
        ).open() as filtered_sources_response,
    ):
        set_up_graphql_mock(m, sources_response)
        results = get_sources()
        assert (
            len(results["source_name"]) == 45
        ), f"Incorrect # of sources: {len(results['name'])}"

        set_up_graphql_mock(m, filtered_sources_response)
        results = get_sources(SourceType.GENE)
        sources = results["source_name"]
        assert len(sources) == 3, f"Incorrect # of sources: {len(sources)}"
        assert set(sources) == {
            "NCBI Gene",
            "HUGO Gene Nomenclature Committee",
            "Ensembl",
        }, "Contains correct sources"


def test_get_gene_list(fixtures_dir: Path, set_up_graphql_mock: Callable):
    with (
        requests_mock.Mocker() as m,
        (
            fixtures_dir
            / "get_gene_list_response.json"  # this fixture is truncated from the real response
        ).open() as gene_list_response,
    ):
        set_up_graphql_mock(m, gene_list_response)

        results = get_all_genes()
        assert len(results["gene_name"]) == 9


def test_get_drug_applications(fixtures_dir, set_up_graphql_mock: Callable):
    with (
        requests_mock.Mocker() as m,
        (
            fixtures_dir / "get_drug_applications_response.json"
        ).open() as drug_applications_response,
        (
            fixtures_dir / "get_drug_applications_drugsatfda_response.json"
        ).open() as drugsatfda_response,
    ):
        set_up_graphql_mock(m, drug_applications_response)
        m.get(
            "https://api.fda.gov/drug/drugsfda.json?search=openfda.application_number:NDA212099&limit=500&skip=0",
            text=drugsatfda_response.read(),
        )
        results = get_drug_applications(["DAROLUTAMIDE"])
        assert len(results["drug_name"]) == 1
        assert results["drug_brand_name"][0] == "NUBEQA"
        assert results["drug_dosage_strength"][0] == "300MG"
        assert (
            results["drug_marketing_status"][0] == ProductMarketingStatus.PRESCRIPTION
        )
        assert results["drug_dosage_form"][0] == ProductDosageForm.TABLET


@pytest.mark.performance()
def test_get_interactions_benchmark(benchmark):
    """Skipped by default -- call pytest with `--performance` flag to run.

    See `conftest.py` for details.
    """
    query = "braf"
    results = benchmark.pedantic(
        get_interactions,
        args=(query,),
        kwargs={"search": "genes"},
        rounds=10,
        warmup_rounds=0,
        iterations=1,
    )
    assert results.columns[0] == "gene"



"""Provides methods for performing different searches in DGIdb"""

import logging
import os
from enum import Enum

import requests
from gql import Client
from gql.transport.requests import RequestsHTTPTransport
from regbot.fetch.drugsfda import get_anda_results, get_nda_results

import dgipy.queries as queries

_logger = logging.getLogger(__name__)

API_ENDPOINT_URL = os.environ.get("DGIDB_API_URL", "https://dgidb.org/api/graphql")


_logger = logging.getLogger(__name__)


def _get_client(api_url: str) -> Client:
    """Acquire GraphQL client.

    :param api_url: endpoint to request data at
    :return: GraphQL client
    """
    transport = RequestsHTTPTransport(
        url=api_url, headers={"dgidb-client-name": "dgipy"}
    )
    return Client(transport=transport, fetch_schema_from_transport=True)


def _group_attributes(row: list[dict]) -> dict:
    grouped_dict = {}
    for attr in row:
        if attr["value"] is None:
            continue
        if attr["name"] in grouped_dict:
            grouped_dict[attr["name"]].append(attr["value"])
        else:
            grouped_dict[attr["name"]] = [attr["value"]]
    return grouped_dict


def _backfill_dicts(col: list[dict]) -> list[dict]:
    keys = {key for cell in col for key in cell}
    return [{key: cell.get(key) for key in keys} for cell in col]


def get_drugs(
    terms: list,
    immunotherapy: bool | None = None,
    antineoplastic: bool | None = None,
    api_url: str | None = None,
) -> dict:
    """Perform a record look up in DGIdb for a drug of interest

    :param terms: drugs for record lookup
    :param immunotherapy: filter option for results that are only immunotherapy
    :param antineoplastic: filter option for results that see antineoplastic use
    :param api_url: API endpoint for GraphQL request
    :return: drug data
    """
    params: dict[str, bool | list] = {"names": terms}
    if immunotherapy is not None:
        params["immunotherapy"] = immunotherapy
    if antineoplastic is not None:
        params["antineoplastic"] = antineoplastic

    api_url = api_url if api_url else API_ENDPOINT_URL
    client = _get_client(api_url)
    result = client.execute(queries.get_drugs.query, variable_values=params)

    output = {
        "drug_name": [],
        "drug_concept_id": [],
        "drug_aliases": [],
        "drug_attributes": [],
        "drug_is_antineoplastic": [],
        "drug_is_immunotherapy": [],
        "drug_is_approved": [],
        "drug_approval_ratings": [],
        "drug_fda_applications": [],
    }
    for match in result["drugs"]["nodes"]:
        output["drug_name"].append(match["name"])
        output["drug_concept_id"].append(match["conceptId"])
        output["drug_aliases"].append([a["alias"] for a in match["drugAliases"]])
        output["drug_attributes"].append(_group_attributes(match["drugAttributes"]))
        output["drug_is_antineoplastic"].append(match["antiNeoplastic"])
        output["drug_is_immunotherapy"].append(match["immunotherapy"])
        output["drug_is_approved"].append(match["approved"])
        output["drug_approval_ratings"].append(
            [
                {"rating": r["rating"], "source": r["source"]["sourceDbName"]}
                for r in match["drugApprovalRatings"]
            ]
        )
        output["drug_fda_applications"].append(
            [app["appNo"] for app in match["drugApplications"]]
        )
    output["drug_attributes"] = _backfill_dicts(output["drug_attributes"])
    return output


def get_genes(terms: list, api_url: str | None = None) -> dict:
    """Perform a record look up in DGIdb for genes of interest

    :param terms: genes for record lookup
    :param api_url: API endpoint for GraphQL request
    :return: gene data
    """
    api_url = api_url if api_url else API_ENDPOINT_URL
    client = _get_client(api_url)
    result = client.execute(queries.get_genes.query, variable_values={"names": terms})

    output = {
        "gene_name": [],
        "gene_concept_id": [],
        "gene_aliases": [],
        "gene_attributes": [],
    }
    for match in result["genes"]["nodes"]:
        output["gene_name"].append(match["name"])
        output["gene_concept_id"].append(match["conceptId"])
        output["gene_aliases"].append([a["alias"] for a in match["geneAliases"]])
        output["gene_attributes"].append(_group_attributes(match["geneAttributes"]))
    output["gene_attributes"] = _backfill_dicts(output["gene_attributes"])
    return output


def get_interactions(
    terms: list,
    search: str = "genes",
    immunotherapy: bool | None = None,
    antineoplastic: bool | None = None,
    source: str | None = None,
    pmid: int | None = None,
    interaction_type: str | None = None,
    approved: str | None = None,
    api_url: str | None = None,
) -> dict:
    """Perform an interaction look up for drugs or genes of interest

    :param terms: drugs or genes for interaction look up
    :param search: interaction search type. valid types are "drugs" or "genes"
    :param immunotherapy: filter option for results that are used in immunotherapy
    :param antineoplastic: filter option for results that are part of antineoplastic regimens
    :param source: filter option for specific database of interest
    :param pmid: filter option for specific PMID
    :param interaction_type: filter option for specific interaction types
    :param approved: filter option for approved interactions
    :param api_url: API endpoint for GraphQL request
    :return: interaction results for terms
    """
    params: dict[str, str | int | bool | list[str]] = {"names": terms}
    if immunotherapy is not None:
        params["immunotherapy"] = immunotherapy
    if antineoplastic is not None:
        params["antiNeoplastic"] = antineoplastic
    if source is not None:
        params["sourceDbName"] = source
    if pmid is not None:
        params["pmid"] = pmid
    if interaction_type is not None:
        params["interactionType"] = interaction_type
    if approved is not None:
        params["approved"] = approved

    api_url = api_url if api_url else API_ENDPOINT_URL
    client = _get_client(api_url)

    if search == "genes":
        raw_results = client.execute(queries.get_interactions_by_gene.query, params)
        results = raw_results["genes"]["nodes"]
    elif search == "drugs":
        raw_results = client.execute(queries.get_interactions_by_drug.query, params)
        results = raw_results["drugs"]["nodes"]
    else:
        msg = "Search type must be specified using: search='drugs' or search='genes'"
        raise Exception(msg)
    output = {
        "gene_name": [],
        "gene_concept_id": [],
        "gene_long_name": [],
        "drug_name": [],
        "drug_concept_id": [],
        "drug_approved": [],
        "interaction_score": [],
        "interaction_attributes": [],
        "interaction_sources": [],
        "interaction_pmids": [],
    }
    for result in results:
        for interaction in result["interactions"]:
            output["gene_name"].append(interaction["gene"]["name"])
            output["gene_long_name"].append(interaction["gene"]["longName"])
            output["gene_concept_id"].append(interaction["gene"]["conceptId"])
            output["drug_name"].append(interaction["drug"]["name"])
            output["drug_concept_id"].append(interaction["drug"]["conceptId"])
            output["drug_approved"].append(interaction["drug"]["approved"])
            output["interaction_score"].append(interaction["interactionScore"])
            output["interaction_attributes"].append(
                _group_attributes(interaction["interactionAttributes"])
            )
            pubs = []
            sources = []
            for claim in interaction["interactionClaims"]:
                sources.append(claim["source"]["sourceDbName"])
                pubs += [p["pmid"] for p in claim["publications"]]
            output["interaction_pmids"].append(pubs)
            output["interaction_sources"].append(sources)
    output["interaction_attributes"] = _backfill_dicts(output["interaction_attributes"])
    return output


def get_categories(terms: list, api_url: str | None = None) -> dict:
    """Perform a category annotation lookup for genes of interest

    :param terms: Genes of interest for annotations
    :param api_url: API endpoint for GraphQL request
    :return: category annotation results for genes
    """
    api_url = api_url if api_url else API_ENDPOINT_URL
    client = _get_client(api_url)
    results = client.execute(
        queries.get_gene_categories.query, variable_values={"names": terms}
    )
    output = {
        "gene_name": [],
        "gene_concept_id": [],
        "gene_full_name": [],
        "gene_category": [],
        "gene_category_sources": [],
    }
    for result in results["genes"]["nodes"]:
        name = result["name"]
        long_name = result["longName"]
        concept_id = result["conceptId"]
        for cat in result["geneCategoriesWithSources"]:
            output["gene_name"].append(name)
            output["gene_concept_id"].append(concept_id)
            output["gene_full_name"].append(long_name)
            output["gene_category"].append(cat["name"])
            output["gene_category_sources"].append(cat["sourceNames"])
    return output


class SourceType(str, Enum):
    """Constrain source types for :py:method:`dgipy.dgidb.get_source` method."""

    DRUG = "drug"
    GENE = "gene"
    INTERACTION = "interaction"
    POTENTIALLY_DRUGGABLE = "potentially_druggable"


def get_sources(
    source_type: SourceType | None = None, api_url: str | None = None
) -> dict:
    """Perform a source lookup for relevant aggregate sources

    >>> from dgipy import get_source, SourceType
    >>> sources = get_source(SourceType.POTENTIALLY_DRUGGABLE)

    :param source_type: type of source to look up. Fetches all sources otherwise.
    :param api_url: API endpoint for GraphQL request
    :return: all sources of relevant type in a json object
    :raise TypeError: if invalid kind of data given as ``source_type`` param.
    """
    source_param = source_type.value.upper() if source_type is not None else None
    api_url = api_url if api_url else API_ENDPOINT_URL
    client = _get_client(api_url)
    params = {} if source_type is None else {"sourceType": source_param}
    results = client.execute(queries.get_sources.query, variable_values=params)
    output = {
        "source_name": [],
        "source_short_name": [],
        "source_version": [],
        "source_drug_claims": [],
        "source_gene_claims": [],
        "source_interaction_claims": [],
        "source_license": [],
        "source_license_url": [],
    }
    for result in results["sources"]["nodes"]:
        output["source_name"].append(result["fullName"])
        output["source_short_name"].append(result["sourceDbName"])
        output["source_version"].append(result["sourceDbVersion"])
        output["source_drug_claims"].append(result["drugClaimsCount"])
        output["source_gene_claims"].append(result["geneClaimsCount"])
        output["source_interaction_claims"].append(result["interactionClaimsCount"])
        output["source_license"].append(result["license"])
        output["source_license_url"].append(result["licenseLink"])
    return output


def get_all_genes(api_url: str | None = None) -> dict:
    """Get all gene names present in DGIdb

    :param api_url: API endpoint for GraphQL request
    :return: list of genes in DGIdb
    """
    api_url = api_url if api_url else API_ENDPOINT_URL
    client = _get_client(api_url)
    results = client.execute(queries.get_all_genes.query)
    genes = {"gene_name": [], "gene_concept_id": []}
    for result in results["genes"]["nodes"]:
        genes["gene_name"].append(result["name"])
        genes["gene_concept_id"].append(result["conceptId"])
    return genes


def get_all_drugs(api_url: str | None = None) -> dict:
    """Get all drug names present in DGIdb

    :param api_url: API endpoint for GraphQL request
    :return: a full list of drugs present in dgidb
    """
    api_url = api_url if api_url else API_ENDPOINT_URL
    client = _get_client(api_url)
    results = client.execute(queries.get_all_drugs.query)
    drugs = {"drug_name": [], "drug_concept_id": []}
    for result in results["drugs"]["nodes"]:
        drugs["drug_name"].append(result["name"])
        drugs["drug_concept_id"].append(result["conceptId"])
    return drugs


def get_drug_applications(terms: list, api_url: str | None = None) -> dict:
    """Perform a look up for ANDA/NDA applications for drug or drugs of interest

    :param terms: drugs of interest
    :param api_url: API endpoint for GraphQL request
    :return: all ANDA/NDA applications for drugs of interest
    """
    api_url = api_url if api_url else API_ENDPOINT_URL
    client = _get_client(api_url)
    results = client.execute(
        queries.get_drug_applications.query, variable_values={"names": terms}
    )
    output = {
        "drug_name": [],
        "drug_concept_id": [],
        "drug_product_application": [],
        "drug_brand_name": [],
        "drug_marketing_status": [],
        "drug_dosage_form": [],
        "drug_dosage_strength": [],
    }

    for result in results["drugs"]["nodes"]:
        name = result["name"]
        concept_id = result["conceptId"]
        for app in result["drugApplications"]:
            app_no = app["appNo"]
            anda = "anda" in app_no
            lui = app_no.split(":")[1]
            full_app_no = f"{'ANDA' if anda else 'NDA'}{lui}"
            try:
                if anda:
                    data = get_anda_results(lui, True)
                else:
                    data = get_nda_results(lui, True)
            except requests.exceptions.RequestException:
                _logger.warning(
                    "HTTP status error for Drugs@FDA lookup %s from drug %s: %s",
                    full_app_no,
                    concept_id,
                    name,
                )
                continue
            if not data:
                _logger.warning(
                    "No results for Drugs@FDA lookup %s from drug %s: %s",
                    full_app_no,
                    concept_id,
                    name,
                )
                continue
            for product in data[0].products:
                output["drug_name"].append(name)
                output["drug_concept_id"].append(concept_id)
                output["drug_product_application"].append(full_app_no)
                output["drug_brand_name"].append(product.brand_name)
                output["drug_marketing_status"].append(product.marketing_status)
                output["drug_dosage_form"].append(product.dosage_form)
                output["drug_dosage_strength"].append(
                    product.active_ingredients[0].strength
                )
    return output



"""Provide utilities relating to data types."""

import csv
from pathlib import Path


def make_tabular(columnar_dict: dict) -> list[dict]:
    """Convert DGIpy query method output to a tabular format.

    :param columnar_dict: column-oriented dict as returned by DGIpy query methods
    :return: list of table rows, where each row keys the column name to the value at
    that column and row.
    """
    return [
        dict(zip(columnar_dict.keys(), row, strict=False))
        for row in zip(*columnar_dict.values(), strict=False)
    ]


def dump_columnar_to_tsv(columnar_dict: dict, output_file: Path) -> None:
    """Dump DGIpy query method output to a TSV file.

    :param columnar_dict: column-oriented dict as returned by DGIpy query methods
    :param output_file: path to save location
    """
    rows = zip(*columnar_dict.values(), strict=False)
    with output_file.open("w", newline="") as tsvfile:
        writer = csv.writer(tsvfile, delimiter="\t")
        writer.writerow(columnar_dict.keys())
        writer.writerows(rows)


def drop_none_attrs(column: list[dict]) -> list[dict]:
    """For an attributes column (i.e., a list of dicts), drop all entries with `None`
    values.

    In DGIdb (and consequently DGIpy), there's no semantic information intended by
    giving attributes `None` values. They are, however, included to ensure compatibility
    with strongly-structured dataframe libraries like Polars. Otherwise, these
    properties are unnecessary, and can be safely dropped without loss of information.

    >>> from dgipy import get_interactions
    >>> from dgipy.data_utils import drop_none_attrs
    >>> results = get_interactions(["braf"])
    >>> results["interaction_attributes"][2]
    {'Response Type': None, 'Combination Therapy': None, 'Novel Drug Target': ['Established target'], 'Variant Effect': None, 'Cancer Type': None, 'Direct Interaction': None, 'Indication': None, 'Clinical Trial Name': ['XL281'], 'Pathway': None, 'Clinical Trial ID': None, 'Mechanism of Action': None, 'Evidence Type': None, 'Alteration': None, 'Approval Status': None}
    >>> results["interaction_attributes"] = drop_none_attrs(
    ...     results["interaction_attributes"]
    ... )
    >>> results["interaction_attributes"][2]
    {'Novel Drug Target': ['Established target'], 'Clinical Trial Name': ['XL281']}

    :param column: an individual column value from the columnar output of a
        :py:module:`dgipy.dgidb` query function
    :return: the same column, but with `None` attributes removed from all cells
    """
    return [{k: v for k, v in d.items() if v is not None} for d in column]
